//
//  AppDelegate.m
//  UpTop
//
//  Created by Geoff Prewett on 11/13/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "AppDelegate.h"

#import "AllProcessesView.h"
#import "Preferences.h"
#import "Settings.h"
#import "TopView.h"

@interface AppDelegate ()
@property NSWindow *window;
@property NSTimer *timer;
@property Model *model;
@property TopView *view;
@property Preferences *prefs;
@property AllProcessesView *allProcs;
@property NSWindow *prefsWindow;
@property NSWindow *allProcsWindow;
@end

@implementation AppDelegate

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification {
    [[NSUserDefaults standardUserDefaults] setObject: [NSNumber numberWithInt: 100] forKey: @"NSInitialToolTipDelay"];
    NSSize screenSize = [NSScreen mainScreen].frame.size;
    NSSize windowSize = NSMakeSize(200, 300);
    NSRect contentSize = NSMakeRect(screenSize.width - windowSize.width, screenSize.height - windowSize.height,
                                    windowSize.width, windowSize.height);
    NSUInteger style = /*NSWindowStyleMaskTitled |*/ NSWindowStyleMaskResizable |
                       NSWindowStyleMaskClosable | NSWindowStyleMaskMiniaturizable;
    self.window = [[NSWindow alloc]
                   initWithContentRect:contentSize styleMask:style backing:NSBackingStoreBuffered defer:YES];
    // Windows can't do blurred transparency without using a private API, but views can.
    // So make the window transparent and do it in the view.
    self.window.opaque = false;
    self.window.backgroundColor = [NSColor colorWithRed:0 green:0 blue:0 alpha:0];
//    self.window.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
    self.window.title = @"Up Top";

    self.model = [[Model alloc] init];
    [self createUI];
    [self.window makeKeyAndOrderFront:self];

    [self update];
    self.timer = nil;  // initialize for -startDelayTimer
    [self startDelayTimer];

    [NSNotificationCenter.defaultCenter
     addObserver:self selector:@selector(startDelayTimer) name:@"DelayChanged" object:nil];
    [NSNotificationCenter.defaultCenter
     addObserver:self selector:@selector(recreateModel) name:@"RecreateModel" object:nil];

    self.prefs = nil;
    self.allProcs = nil;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification {
    [self.timer invalidate];  // stop timer;  might be unnecessary
    if (Settings.windowLocation >= kLocMoveable) {
        [Settings setDefaultWindowRect:self.window.frame];
    }
}

- (void)update {
    [self.model update];
    [self.view update];
    if (self.allProcs != nil) {
        [self.allProcs update];
    }
}

- (void)createUI
{
    self.view = [[TopView alloc] initWithFrame:CGRectZero];
    self.view.model = self.model;
    [self.view update];
    [self.view sizeToFit];
    // Resize window (note:  need to do this before putting the view in, since it
    //                       gets resized to the window's size)
    NSSize screenSize = [NSScreen mainScreen].frame.size;
    NSSize windowSize = self.view.frame.size;
    if (Settings.windowLocation >= kLocMoveable) {
        self.window.styleMask |= NSWindowStyleMaskTitled;
        windowSize = [self.window frameRectForContentRect:NSMakeRect(0, 0, windowSize.width, windowSize.height)].size;
    } else {
        if (self.window.styleMask & NSWindowStyleMaskTitled) {  // not just starting
            [Settings setDefaultWindowRect:self.window.frame];
        }
        self.window.styleMask &= ~NSWindowStyleMaskTitled;
    }
    NSRect contentSize;
    switch (Settings.windowLocation) {
        case kLocUpperRight:
            contentSize = NSMakeRect(screenSize.width - windowSize.width, screenSize.height - windowSize.height,
                                     windowSize.width, windowSize.height);
            break;
        case kLocMiddleRight:
            contentSize = NSMakeRect(screenSize.width - windowSize.width,
                                     (screenSize.height - windowSize.height) / 2.0,
                                     windowSize.width, windowSize.height);
            break;
        case kLocLowerRight:
            contentSize = NSMakeRect(screenSize.width - windowSize.width, 0,
                                     windowSize.width, windowSize.height);
            break;
        case kLocUpperLeft:
            contentSize = NSMakeRect(0, screenSize.height - windowSize.height,
                                     windowSize.width, windowSize.height);
            break;
        case kLocMiddleLeft:
            contentSize = NSMakeRect(0, (screenSize.height - windowSize.height) / 2.0,
                                     windowSize.width, windowSize.height);
            break;
        case kLocLowerLeft:
            contentSize = NSMakeRect(0, 0, windowSize.width, windowSize.height);
            break;
        default: {
            NSRect r = Settings.windowRect;  // ignore size, in case it changed
            contentSize = NSMakeRect(NSMinX(r), NSMinY(r), windowSize.width, windowSize.height);
            }
            break;
    }
    [self.window setFrame:contentSize display:YES animate:NO];

    self.window.contentView = self.view;
}

- (void)startDelayTimer
{
    if (self.timer) {
        [self.timer invalidate];
    }
    self.timer = [NSTimer scheduledTimerWithTimeInterval:Settings.refreshDelaySec repeats:YES block:^(NSTimer *timer) {
        [self update];
    }];
    self.timer.tolerance = Settings.refreshDelaySec * .05;  // Apple says this helps power usage
}

- (void)recreateModel
{
    self.model = [[Model alloc] init];
    self.view.model = self.model;
}

- (IBAction)onPreferences:(id)sender
{
    if (self.prefs == nil) {
        self.prefs = [[Preferences alloc] initWithFrame:CGRectZero];
        [self.prefs sizeToFit];
        NSSize windowSize = self.prefs.frame.size;
        [NSNotificationCenter.defaultCenter
         addObserver:self selector:@selector(createUI) name:@"UpdateUI" object:nil];

        NSSize screenSize = [NSScreen mainScreen].frame.size;
        NSRect contentSize = NSMakeRect((screenSize.width - windowSize.width) / 2,
                                        (screenSize.height - windowSize.height) / 2,
                                        windowSize.width, windowSize.height);
        NSUInteger style = NSWindowStyleMaskTitled | NSWindowStyleMaskResizable | NSWindowStyleMaskClosable;
        self.prefsWindow = [[NSWindow alloc]
                            initWithContentRect:contentSize styleMask:style backing:NSBackingStoreBuffered defer:YES];
        self.prefsWindow.releasedWhenClosed = NO;
        self.prefsWindow.contentView = self.prefs;
        self.prefsWindow.title = @"UpTop Preferences";
    }
    [self.prefsWindow makeKeyAndOrderFront:self];
}

- (IBAction)onShowAllProcesses:(id)sender
{
    if (self.allProcs == nil) {
        self.allProcs = [[AllProcessesView alloc] initWithFrame:NSMakeRect(0, 0, 1000, 400)];
        self.allProcs.model = self.model;
        [self.allProcs update];
        NSSize windowSize = self.allProcs.frame.size;
        NSSize screenSize = [NSScreen mainScreen].frame.size;
        NSRect contentSize = NSMakeRect((screenSize.width - windowSize.width) / 2,
                                        (screenSize.height - windowSize.height) / 2,
                                        windowSize.width, windowSize.height);
        NSUInteger style = NSWindowStyleMaskTitled | NSWindowStyleMaskResizable | NSWindowStyleMaskClosable;
        self.allProcsWindow = [[NSWindow alloc]
                            initWithContentRect:contentSize styleMask:style backing:NSBackingStoreBuffered defer:YES];
        self.allProcsWindow.releasedWhenClosed = NO;
        self.allProcsWindow.contentView = self.allProcs;
        self.allProcsWindow.title = @"UpTop: All Processes";
    }
    [self.allProcsWindow makeKeyAndOrderFront:self];
}
@end
