//
//  BarGraph.m
//  UpTop
//
//  Created by Geoff Prewett on 11/13/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "BarGraph.h"

@interface BarGraph ()
{
    double mTotal;
    NSArray<NSNumber*>* mAmounts;
    NSArray<NSColor*>* mColors;
    NSArray<NSString*>* mTooltips;
}
@end

@implementation BarGraph
- (instancetype)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        self.colors = @[[NSColor redColor], [NSColor greenColor], [NSColor blueColor]];
    }
    return self;
}

- (NSArray<NSColor *> *)colors
{
    return mColors;
}

- (void)setColors:(NSArray<NSColor *> *)colors
{
    mColors = colors;
    self.needsDisplay = YES;
}

- (double)total
{
    return mTotal;
}

- (void)setTotal:(double)total
{
    mTotal = total;
    self.needsDisplay = YES;
}

- (NSArray<NSNumber *> *)amounts
{
    return mAmounts;
}

- (void)setAmounts:(NSArray<NSNumber *> *)amounts
{
    mAmounts = amounts;
    self.needsDisplay = YES;
}

- (NSArray<NSString *> *)tooltips
{
    return mTooltips;
}

- (void)setTooltips:(NSArray<NSString *> *)tooltips
{
    mTooltips = tooltips;
    self.needsDisplay = YES;
}

- (void)drawRect:(NSRect)dirtyRect {
    [super drawRect:dirtyRect];

    [self removeAllToolTips];

    NSGraphicsContext *gc = [NSGraphicsContext currentContext];
    CGContextRef context = (CGContextRef)[gc graphicsPort];

    int idx = 0;
    int colorIdx = 0;
    CGFloat width = self.frame.size.width;
    CGFloat height = self.frame.size.height;
    CGFloat x = 0;
    for (NSNumber* n in self.amounts) {
        CGFloat w = (n.doubleValue / self.total) * width;
        w = MAX(1.0, w);
        CGContextSetFillColorWithColor(context, [self.colors objectAtIndex:colorIdx].CGColor);
        CGContextFillRect(context, NSMakeRect(x, 0, w, height));

        if (self.tooltips.count > idx) {
            [self addToolTipRect:NSMakeRect(x, 0, w, height) owner:[self.tooltips objectAtIndex:idx] userData:nil];
        }

        x += w;
        colorIdx++;
        if (colorIdx >= self.colors.count) {
            colorIdx = 0;
        }
        idx++;

    }
}

@end
