//
//  FontPicker.m
//  UpTop
//
//  Created by Geoff Prewett on 11/25/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "FontPicker.h"

#define kSystemFontName     @"[System Font]"

@interface FontPicker ()
@property NSPopUpButton* popup;
@property NSString* systemFontName;
@end

@implementation FontPicker

- (instancetype)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        self.systemFontName = self.popup.font.familyName;  // save, in case user changes control font (admittedly unlikely)
        self.popup = [[NSPopUpButton alloc] initWithFrame:frame pullsDown:NO];
        self.popup.target = self;
        self.popup.action = @selector(onChanged:);
        [self addSubview:self.popup];
        [self update];
        [self.popup selectItemWithTitle:kSystemFontName];
    }
    return self;
}

- (BOOL)isSystemFont
{
    return [self.popup.selectedItem.title isEqualToString:kSystemFontName];
}

- (NSString*)selectedFontName
{
    if (self.isSystemFont == YES) {
        return self.systemFontName;
    }
    return self.popup.selectedItem.title;
}

- (void)setSelectedFontName:(NSString *)fontName
{
    NSString *current = self.popup.selectedItem.title;  // in case fontName isn't valid, we can restore
    if ([fontName isEqualToString:self.systemFontName]) {
        [self.popup selectItemWithTitle:kSystemFontName];
    } else {
        [self.popup selectItemWithTitle:fontName];
    }
    if (self.popup.selectedItem == nil) {
        [self.popup selectItemWithTitle:current];
    }
}

- (void)update
{
    [self.popup removeAllItems];
    [self.popup addItemWithTitle:kSystemFontName];
    for (NSString *name in NSFontManager.sharedFontManager.availableFontFamilies) {
        NSFont *f = [NSFont fontWithName:name size:12];
        if (f.mostCompatibleStringEncoding <= NSMacOSRomanStringEncoding
            && f.numberOfGlyphs > 150) {  // eliminates STX... number-only fonts
            [self.popup addItemWithTitle:name];
        }
    }
}

- (void)sizeToFit
{
    [super sizeToFit];
    [self.popup sizeToFit];

    NSDictionary *attrs = @{ NSFontAttributeName: self.popup.font };
    CGFloat maxWidth = 0;
    for (NSMenuItem *mi in self.popup.itemArray) {
        maxWidth = MAX(maxWidth, [mi.title sizeWithAttributes:attrs].width);
    }
    CGFloat w = maxWidth + self.frame.size.height;  // account for pull-down button on right side
    self.popup.frame = NSMakeRect(0, 0, w, self.popup.frame.size.height);
    self.frame = NSMakeRect(self.frame.origin.x, self.frame.origin.y,
                            self.popup.frame.size.width, self.popup.frame.size.height);
}

- (void)onChanged:(id)sender
{
    [self sendAction:self.action to:self.target];
}
@end
