//
//  Model.h
//  UpTop
//
//  Created by Geoff Prewett on 11/25/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import <Foundation/Foundation.h>

#import "Accumulator.h"
#import "Process.h"

/*
 Per-process CPU percentages are inherently inaccurate.  Hyperthreading allows things to execute in
 parallel--usually.  So we divide the CPU usage by the number of logical cores, but some things can't
 execute in parallel, which will throw the numbers off.
 */

@interface Model : NSObject
@property (readonly) NSString* error;
@property (readonly) int nCores;

@property (readonly) double cpuSystemPercentage;
@property (readonly) double cpuUserPercentage;
@property (readonly) double cpuIdlePercentage;
@property (readonly) BOOL isCPUvalid;  // won't be valid on first tick

@property (readonly) double memoryTotal;
@property (readonly) double memoryUsed;
@property (readonly) double memoryWired;
@property (readonly) double memoryCompressed;
@property (readonly) double memoryApps;
@property (readonly) double swapTotal;
@property (readonly) double swapUsed;

// Total bytes read/written to all disks during this sample 
@property (readonly) uint64_t diskBytesRead;
@property (readonly) uint64_t diskBytesWritten;

@property (readonly) NSArray* processes; // all the processes existing in this sample
@property (readonly) NSArray* diskPocesses;  // processes that performed disk I/O during this sample

- (instancetype)init;
- (void)update;
@end
