//
//  Model.m
//  UpTop
//
//  Created by Geoff Prewett on 11/25/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "Model.h"

#import "Settings.h"
#import "Utils.h"

#include <sys/sysctl.h>
#include <mach/host_info.h>
#include <mach/mach_host.h>
#include <mach/task_info.h>
#include <mach/task.h>
//#include <mach/shared_memory_server.h>
#include <mach/shared_region.h>

#include <sys/param.h>
#include <sys/sysctl.h>
#include <pwd.h>

#include <libproc.h>

#include <Security/Authorization.h>

typedef struct {
    uint64_t system;
    uint64_t user;
    uint64_t idle;
} CPUSample;

typedef struct kinfo_proc kinfo_proc;

int getNCores();
int getBSDProcessList(kinfo_proc **procList, size_t *procCount);
NSArray* getProcessList();

@interface Model ()
{
    int mPageSize;
    CPUSample mLastCPU;
    CPUSample mThisCPU;
    double mLastTime;
}
@property NSDictionary* lastProcs;
@property NSDictionary* thisProcs;
@property Accumulator* diskProcs;

@end

@implementation Model
- (instancetype)init
{
    self = [super init];
    if (self) {
        mPageSize = [self pagesize];
        mLastCPU.system = -1;
        mLastCPU.user = -1;
        mLastCPU.idle = -1;
        mThisCPU = mLastCPU;
        self.lastProcs = nil;
        self.thisProcs = nil;
        [self createDiskAccumulator];
//        self.diskProcs = nil;

        self->_nCores = getNCores();

        self->_cpuSystemPercentage = 100;
        self->_cpuUserPercentage = 0;
        self->_cpuIdlePercentage = 0;
        self->_isCPUvalid = NO;

        self->_memoryTotal = 0;
        self->_memoryUsed = 0;
        self->_memoryWired = 0;
        self->_memoryCompressed = 0;
        self->_memoryApps = 0;

        self->_diskBytesRead = 0;
        self->_diskBytesWritten = 0;
    }
    return self;
}

- (void)createDiskAccumulator
{
    int len = Settings.diskIOAccumulatorLength;
    self.diskProcs = [[Accumulator alloc] initWithSize:len];
}

- (int)pagesize
{
    int mib[6];
    mib[0] = CTL_HW;
    mib[1] = HW_PAGESIZE;
    
    int pagesize;
    size_t length;
    length = sizeof(pagesize);
    if (sysctl(mib, 2, &pagesize, &length, NULL, 0) < 0)
    {
        self->_error = @"Could not get page size:  sysctl([CTL_HW, HW_PAGESIZE]) failed";
    }
    return pagesize;
}

- (BOOL)acquireTaskportRight
{
    // From http://os-tres.net/blog/2010/02/17/mac-os-x-and-task-for-pid-mach-call/
    OSStatus stat;
    AuthorizationItem taskport_item[] = {{"system.privilege.taskport:"}};
    AuthorizationRights rights = {1, taskport_item}, *out_rights = NULL;
    AuthorizationRef author;
    
    AuthorizationFlags auth_flags = kAuthorizationFlagExtendRights | kAuthorizationFlagPreAuthorize | kAuthorizationFlagInteractionAllowed | ( 1 << 5);
    
    stat = AuthorizationCreate (NULL, kAuthorizationEmptyEnvironment,auth_flags, &author);
    if (stat != errAuthorizationSuccess)
    {
        return NO;
    }
    
    stat = AuthorizationCopyRights ( author, &rights, kAuthorizationEmptyEnvironment, auth_flags, &out_rights);
    if (stat != errAuthorizationSuccess)
    {
        return NO;
    }
    return YES;
}


- (void)update
{
    self->_error = nil;
    if (self.nCores <= 0) {  // shouldn't actually ever be zero
        self->_error = @"Could not get number of CPU cores:  sysctl([CTL_HW, HW_PAGESIZE], 2) failed";
    }
    struct timeval t;
    gettimeofday(&t, NULL);
    double now = (double)t.tv_sec + (double)t.tv_usec / 1000000.0;
    double dt = now - mLastTime;
    
    [self updateCPU];
    [self updateMemory];
    
    NSArray *procs = getProcessList();
    
    BOOL isFirstTick = (self.thisProcs == nil);
    if (!isFirstTick) {
        self.lastProcs = self.thisProcs;
    }
    uint64_t totalRead = 0;
    uint64_t totalWritten = 0;
    uint64_t totalCPU = 0;
    NSMutableDictionary *procDict = [[NSMutableDictionary alloc] initWithCapacity:procs.count];
    for (Process *p in procs) {
        Process *oldP = [self.lastProcs objectForKey:@(p.processID)];
        totalCPU += (p.totalCPUTimeRaw - oldP.totalCPUTimeRaw);
    }
    for (Process *p in procs) {
        Process *oldP = [self.lastProcs objectForKey:@(p.processID)];
        p.diskRead = p.totalDiskRead - oldP.totalDiskRead;
        p.diskWritten = p.totalDiskWritten - oldP.totalDiskWritten;
//        p.cpuPercent = (double)(p.totalCPUTimeRaw - oldP.totalCPUTimeRaw) / (double)totalCPU * 100.0 / (double)self.nCores;
        p.cpuPercent = (double)(p.totalCPUTimeRaw - oldP.totalCPUTimeRaw) / 1e9 / dt * 100.0;
        totalRead += p.diskRead;
        totalWritten += p.diskWritten;
        [procDict setObject:p forKey:@(p.processID)];
    }

/*    uint64_t dSystem = mThisCPU.system - mLastCPU.system;
     uint64_t dUser = mThisCPU.user - mLastCPU.user;
     uint64_t dIdle = mThisCPU.idle - mLastCPU.idle;
     // Add the deltas to avoid overflow
     uint64_t dTotal = dSystem + dUser + dIdle;
//     printf("total ticks: %d, used ticks: %d, added: %.1f, dt: %.1fs %.1f%% %.1f%%\n", dTotal, dSystem + dUser, totalCPU / 10000000.0, dt, ((double)totalCPU / 1000000000.0) / dt * 25.0, (double)(dSystem + dUser) / (double)dTotal * 100.0);
    printf("used: %.1f%%, used_t: %.1fs, dt: %.1fs, %.1f%%\n", (double)(dSystem + dUser) / (double)dTotal * 100.0, totalCPU / 10000000.0, dt, ((double)totalCPU / 1000000000.0) / dt * 25.0 ); */
    self.thisProcs = procDict;
    self->_diskBytesRead = totalRead;
    self->_diskBytesWritten = totalWritten;
    self->_processes = procs;

    NSArray *diskSortedProcs = [self sortByDiskIO:procs];
    if (self.diskProcs != nil) {
        if (isFirstTick == YES) {
            diskSortedProcs = @[];
        } else {
            [self.diskProcs addObject:diskSortedProcs];
            NSMutableDictionary *ioProcs = [[NSMutableDictionary alloc] init];
            for (NSArray *a in self.diskProcs.data) {
                for (Process *p in a) {
                    Process *accumulated = [ioProcs objectForKey:@(p.processID)];
                    if (accumulated) {
                        accumulated.diskRead += p.diskRead;
                        accumulated.diskWritten += p.diskWritten;
                    } else {
                        [ioProcs setObject:p forKey:@(p.processID)];
                    }
                }
            }
            diskSortedProcs = [self sortByDiskIO:[ioProcs allValues]];
        }
    }
    self->_diskPocesses = diskSortedProcs;

    mLastTime = now;
}

- (NSArray*)sortByDiskIO:(NSArray *)procs
{
    NSSortDescriptor *sort = [[NSSortDescriptor alloc] initWithKey:@"diskWritten" ascending:NO];
    NSArray *writesSortedProcs = [procs sortedArrayUsingDescriptors:@[sort]];
    sort = [[NSSortDescriptor alloc] initWithKey:@"diskRead" ascending:NO];
    NSArray *readsSortedProcs = [procs sortedArrayUsingDescriptors:@[sort]];
    NSMutableArray *diskSortedProcs = [[NSMutableArray alloc] init];
    NSMutableSet *alreadyAdded = [[NSMutableSet alloc] init];
    for (Process *p in writesSortedProcs) {
        if (p.diskWritten == 0) {
            break;
        }
        [diskSortedProcs addObject:p];
        [alreadyAdded addObject:@(p.processID)];
    }
    for (Process *p in readsSortedProcs) {
        if (p.diskRead == 0) {
            break;
        }
        if ([alreadyAdded containsObject:@(p.processID)] == NO) {
            [diskSortedProcs addObject:p];
        }
    }
    return diskSortedProcs;
}

- (void)updateCPU
{
    kern_return_t ret;
    mach_msg_type_number_t count;
    host_cpu_load_info_data_t load;
    
    count = HOST_CPU_LOAD_INFO_COUNT;
    ret = host_statistics(mach_host_self(), HOST_CPU_LOAD_INFO, (int *)&load, &count);
    if (ret == KERN_SUCCESS) {
        mLastCPU = mThisCPU;
        mThisCPU.system = load.cpu_ticks[CPU_STATE_SYSTEM];
        mThisCPU.user = load.cpu_ticks[CPU_STATE_USER] + load.cpu_ticks[CPU_STATE_NICE];
        mThisCPU.idle = load.cpu_ticks[CPU_STATE_IDLE];
        // Guard against rollovers
        if (mThisCPU.system >= mLastCPU.system && mThisCPU.user >= mLastCPU.user && mThisCPU.idle >= mLastCPU.idle) {
            uint64_t dSystem = mThisCPU.system - mLastCPU.system;
            uint64_t dUser = mThisCPU.user - mLastCPU.user;
            uint64_t dIdle = mThisCPU.idle - mLastCPU.idle;
            // Add the deltas to avoid overflow
            uint64_t dTotal = dSystem + dUser + dIdle;
            if (dTotal == 0) { dTotal = 1; } // not likely...
            self->_cpuSystemPercentage = (double)dSystem / (double)dTotal * 100.0;
            self->_cpuUserPercentage = (double)dUser / (double)dTotal * 100.0;
            self->_cpuIdlePercentage = (double)(dTotal - (dSystem + dUser)) / (double)dTotal * 100.0;
            self->_isCPUvalid = YES;
        } else {
            self->_cpuSystemPercentage = 100;
            self->_cpuUserPercentage = 0;
            self->_cpuIdlePercentage = 0;
            self->_isCPUvalid = NO;
        }
    }
}

- (void)updateMemory
{
/*    int mib[6];
    mib[0] = CTL_HW;
    mib[1] = HW_PAGESIZE;
    
    int pagesize;
    size_t length;
    length = sizeof(pagesize);
    if (sysctl(mib, 2, &pagesize, &length, NULL, 0) < 0)
    {
        self->_error = @"Could not get page size:  sysctl([CTL_HW, HW_PAGESIZE]) failed";
    } */
    size_t length;
    int pagesize = self->mPageSize;

    struct xsw_usage swapInfo;
    length = sizeof(swapInfo);
    if (sysctlbyname("vm.swapusage", &swapInfo, &length, NULL, 0) != KERN_SUCCESS) {
        self->_error = @"Could not get swap usage:  sysctlbyname('vm.swapusage') failed";
    }

    mach_msg_type_number_t count = HOST_VM_INFO64_COUNT;
    vm_statistics64_data_t vmstat;
    if (host_statistics64(mach_host_self(), HOST_VM_INFO64, (host_info_t)&vmstat, &count) != KERN_SUCCESS)
    {
        self->_error = @"Could not get VM stats:  host_statistics64() failed";
    }
    
    double wired = (double)vmstat.wire_count * (double)pagesize;
    double active = (double)vmstat.active_count * (double)pagesize;
    double inactive = (double)vmstat.inactive_count * (double)pagesize;
    double speculative  = (double)vmstat.speculative_count * (double)pagesize;
    double free = (double)vmstat.free_count * (double)pagesize;
    double compressed = (double)vmstat.compressor_page_count * (double)pagesize;
    double purgeable = (double)vmstat.purgeable_count * (double)pagesize;
    double internal = (double)vmstat.internal_page_count * (double)pagesize;
    double external = (double)vmstat.external_page_count * (double)pagesize;
    double uncompressed = (double)vmstat.total_uncompressed_pages_in_compressor * (double)pagesize;
    double swapTotal = (double)swapInfo.xsu_total;
    double swapUsed = (double)swapInfo.xsu_used;

    double appMemory = internal - purgeable;
    double used = appMemory + wired + compressed;
    double cachedFiles = external + purgeable;
    double total = used + cachedFiles + free;

/*    double gb = 1024.0 * 1024.0 * 1024.0;
    printf("Wired: %.2f GB\n", wired / gb);
    printf("Active: %.2f GB\n", active / gb);
    printf("Inactive: %.2f GB\n", inactive / gb);
    printf("Speculative: %.2f GB\n", speculative / gb);
    printf("Free: %.2f GB\n", free / gb);
    printf("Compressed: %.2f GB\n", compressed / gb);
    printf("Purgeable: %.2f GB\n", purgeable / gb);
    printf("Internal: %.2f GB\n", internal / gb);
    printf("External: %.2f GB\n", external / gb);
    printf("Uncompressed in compressor: %.2f GB\n", uncompressed / gb);
    printf("Swap used: %.2f GB\n", swapUsed / gb);
    printf("\n");
    printf("App memory: %.2f GB\n", appMemory / gb);
    printf("Cached files: %.2f GB\n", cachedFiles / gb);
    printf("Total: %.2f GB\n", total / gb); */

    self->_memoryTotal = total;
    self->_memoryUsed = used;
    self->_memoryWired = wired;
    self->_memoryCompressed = compressed;
    self->_memoryApps = appMemory;
    self->_swapTotal = swapTotal;
    self->_swapUsed = swapUsed;
}

@end

int getNCores()
{
    // Physical cores would be "hw.physicalcpu", but we don't want that, because of hyperthreading
    int nCores = -1;
    size_t sizeOf = sizeof(nCores);
    sysctlbyname("hw.logicalcpu", &nCores, &sizeOf, NULL, 0);
    return nCores;
}

// From:  http://stackoverflow.com/questions/18820199/unable-to-detect-application-running-with-another-user-via-switch-user/18821357#18821357
int getBSDProcessList(kinfo_proc **procList, size_t *procCount)
// Returns a list of all BSD processes on the system.  This routine
// allocates the list and puts it in *procList and a count of the
// number of entries in *procCount.  You are responsible for freeing
// this list (use "free" from System framework).
// On success, the function returns 0.
// On error, the function returns a BSD errno value.
{
    int                 err;
    kinfo_proc *        result;
    bool                done;
    static const int    name[] = { CTL_KERN, KERN_PROC, KERN_PROC_ALL, 0 };
    // Declaring name as const requires us to cast it when passing it to
    // sysctl because the prototype doesn't include the const modifier.
    size_t              length;
    
    *procCount = 0;
    
    // We start by calling sysctl with result == NULL and length == 0.
    // That will succeed, and set length to the appropriate length.
    // We then allocate a buffer of that size and call sysctl again
    // with that buffer.  If that succeeds, we're done.  If that fails
    // with ENOMEM, we have to throw away our buffer and loop.  Note
    // that the loop causes us to call sysctl with NULL again; this
    // is necessary because the ENOMEM failure case sets length to
    // the amount of data returned, not the amount of data that
    // could have been returned.
    result = NULL;
    done = false;
    do {
        assert(result == NULL);
        
        // Call sysctl with a NULL buffer.
        length = 0;
        err = sysctl((int *)name, (sizeof(name) / sizeof(*name)) - 1,
                     NULL, &length,
                     NULL, 0);
        if (err == -1) {
            err = errno;
        }
        
        // Allocate an appropriately sized buffer based on the results
        // from the previous call.
        if (err == 0) {
            result = malloc(length);
            if (result == NULL) {
                err = ENOMEM;
            }
        }
        
        // Call sysctl again with the new buffer.  If we get an ENOMEM
        // error, toss away our buffer and start again.
        if (err == 0) {
            err = sysctl((int *)name, (sizeof(name) / sizeof(*name)) - 1,
                         result, &length,
                         NULL, 0);
            if (err == -1) {
                err = errno;
            }
            if (err == 0) {
                done = true;
            } else if (err == ENOMEM) {
                assert(result != NULL);
                free(result);
                result = NULL;
                err = 0;
            }
        }
    } while (err == 0 && ! done);
    
    // Clean up and establish post conditions
    if (err != 0 && result != NULL) {
        free(result);
        result = NULL;
    }
    *procList = result;
    if (err == 0) {
        *procCount = length / sizeof(kinfo_proc);
    }
    
    assert( (err == 0) == (*procList != NULL) );
    
    return err;
}

NSArray* getProcessList()
{
    kinfo_proc *procList =NULL;
    size_t nProcs = 0;
    getBSDProcessList(&procList, &nProcs);
    
    NSMutableArray *processes = [NSMutableArray arrayWithCapacity:(int)nProcs];
    
    for (int i = 0; i < nProcs; i++) {
        struct kinfo_proc *currentProcess = &procList[i];
        struct passwd *user = getpwuid(currentProcess->kp_eproc.e_ucred.cr_uid);
        Process *proc = [[Process alloc] init];
        proc.processID = currentProcess->kp_proc.p_pid;
        proc.processName = [NSString stringWithFormat: @"%s",currentProcess->kp_proc.p_comm];
        NSRunningApplication *app = [NSRunningApplication runningApplicationWithProcessIdentifier:proc.processID];
        if (app) {
            proc.localizedProcessName = app.localizedName;
        } else { 
            proc.localizedProcessName = @"";
        }
        proc.userID = currentProcess->kp_eproc.e_ucred.cr_uid;
        // TODO:  it would be more efficient to cache the NSStrings for username
        proc.userName = [NSString stringWithFormat: @"%s",user->pw_name];
        proc.cpuPercent = 0.0;
        proc.totalCPUTime = 0.0;

        char pathbuf[PROC_PIDPATHINFO_MAXSIZE];
        if (proc_pidpath(currentProcess->kp_proc.p_pid, pathbuf, sizeof(pathbuf)) > 0) {
            proc.processPath = [NSString stringWithUTF8String:pathbuf];
            // kp_proc.p_comm can get truncated, so fix up proc.processName
            int lastDirIdx = -1;
            char *p = pathbuf;
            while (*p != '\0' && p < pathbuf + sizeof(pathbuf)) { // paranoia with the size check
                if (*p == '/') {
                    lastDirIdx = (int)(p - pathbuf);
                }
                ++p;
            }
            if (lastDirIdx >= 0) {
                proc.processName = [NSString stringWithUTF8String:(pathbuf + lastDirIdx + 1)];
            }
        }

/*        // Process information is based on Apple's ps implementation
        // https://opensource.apple.com/source/adv_cmds/adv_cmds-163/ps/tasks.c.auto.html
        task_port_t task;
        kern_return_t ret = task_for_pid(mach_task_self(), proc.processID, &task);
        if (ret != KERN_SUCCESS) {
            //            printf("task_for_pid(%d {%s}) failed with %d\n", proc.processID, proc.processName.UTF8String, ret);
        } else {
            //            printf("task_for_pid(%d {%s}) success!\n", proc.processID, proc.processName.UTF8String);
            struct mach_task_basic_info info;
            unsigned sizeOf = sizeof(info);
            ret = task_info(task, MACH_TASK_BASIC_INFO, (task_info_t)&info, &sizeOf);
            if (ret == KERN_SUCCESS) {
                proc.virtualMem = info.virtual_size;
                proc.residentMem = info.resident_size;
                
                unsigned int nThreads;
                thread_port_array_t threads;
                struct thread_basic_info threadVal;
                ret = task_threads(task, &threads, &nThreads);
                if (ret == KERN_SUCCESS) {
                    integer_t cpuPercentScaled = 0.0;
                    sizeOf = THREAD_BASIC_INFO_COUNT;
                    for (int i = 0;  i < nThreads;  ++i) {
                        ret = thread_info(threads[i], THREAD_BASIC_INFO, (thread_info_t)&threadVal, &sizeOf);
                        if (ret == KERN_SUCCESS) {
                            cpuPercentScaled += threadVal.cpu_usage;
                        }
                        mach_port_deallocate(mach_task_self(), threads[i]);
                    }
                    // Heh, turns out cpu_usage is a scaled percentage.  And not
                    // even a percentage, a ratio.
                    proc.cpuPercent = (double)cpuPercentScaled / (double)TH_USAGE_SCALE * 100.0;
                }
            }
            mach_port_deallocate(mach_task_self(), task);
        } */
        
        struct rusage_info_v3 usage;
        if (proc_pid_rusage(proc.processID, RUSAGE_INFO_V3, (rusage_info_t*)&usage) == 0) {
//            printf("%d: read: %llu, written: %llu\n", proc.processID, usage.ri_diskio_bytesread,
//                   usage.ri_diskio_byteswritten);
/*            printf("%s: footprint: %.1f MB, wired: %.1f, res: %.1f\n", proc.processName.UTF8String,
                   toMB(usage.ri_phys_footprint),
                   toMB(usage.ri_wired_size),
                   toMB(usage.ri_resident_size)); */
            proc.totalCPUTimeRaw = usage.ri_system_time + usage.ri_user_time;
            proc.totalDiskRead = usage.ri_diskio_bytesread;
            proc.totalDiskWritten = usage.ri_diskio_byteswritten;
            // usage.ri_resident_size does not seem to be reliable;  frequently it reports more memory
            // than usage.ri_phys_footprint.  So, no RSS for us.
            proc.totalMem = usage.ri_phys_footprint;
        }
        
        [processes addObject:proc];
    }
    free(procList);
    
    return processes;
}
