//
//  NumericEdit.m
//  UpTop
//
//  Created by Geoff Prewett on 11/25/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "NumericEdit.h"

@interface NumericEdit ()
{
    double mMinValue;
    double mMaxValue;
    double mInc;
}
@property NSTextField* text;
@property NSStepper* stepper;
@end

@implementation NumericEdit

- (instancetype)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        self.text = [[NSTextField alloc] initWithFrame:CGRectZero];
        self.text.delegate = self;
        [self.text sizeToFit];
        [self addSubview:self.text];

        self.stepper = [[NSStepper alloc] initWithFrame:CGRectZero];
        self.stepper.target = self;
        self.stepper.action = @selector(onStepped:);
        self.stepper.valueWraps = NO;
        [self.stepper sizeToFit];
        [self addSubview:self.stepper];

        self.minValue = 0.0;
        self.maxValue = 100.0;
        self.increment = 1.0;
        self.doubleValue = 0.0;
    }
    return self;
}

- (void)sizeToFit
{
    [self.text sizeToFit];
    [self.stepper sizeToFit];
    // Note:  it might be tempting to try to make the height of both the largest height of either,
    //        but it turns out that the stepper is bigger, and then the text field is too big and
    //        looks awkward.
    CGFloat h = self.text.frame.size.height;
    self.text.frame = NSMakeRect(0, 0, MAX(1.5 * h, self.text.frame.size.width), h);
    self.frame = NSMakeRect(self.frame.origin.x, self.frame.origin.y,
                            self.text.frame.size.width + self.stepper.frame.size.width,
                            h);
}

- (void)updateText
{
    self.text.stringValue = [NSString stringWithFormat:@"%g", self.doubleValue];
}

- (void)onStepped:(id)sender
{
    [self updateText];
    if (self.target != nil && self.action != nil) {
        [self sendAction:self.action to:self.target];
    }
}

- (void)setFrame:(NSRect)frame
{
    [super setFrame:frame];

    self.stepper.frame = NSMakeRect(frame.size.width - self.stepper.frame.size.width,
                                    (frame.size.height - self.stepper.frame.size.height) / 2,
                                    self.stepper.frame.size.width,
                                    frame.size.height);
    self.text.frame = NSMakeRect(0, (frame.size.height - self.text.frame.size.height) / 2,
                                 frame.size.width - self.stepper.frame.size.width,
                                 frame.size.height);
}

- (double)minValue
{
    return mMinValue;
}

- (void)setMinValue:(double)minValue
{
    if (minValue >= mMaxValue) {
        return;
    }
    mMinValue = minValue;
    self.stepper.minValue = minValue;
}

- (double)maxValue
{
    return mMaxValue;
}

- (void)setMaxValue:(double)maxValue
{
    if (maxValue <= mMinValue) {
        return;
    }
    mMaxValue = maxValue;
    self.stepper.maxValue = maxValue;
}

- (double)increment
{
    return mInc;
}

- (void)setIncrement:(double)inc
{
    mInc = inc;
    self.stepper.increment = inc;
}

- (double)doubleValue
{
    return self.stepper.doubleValue;
}

- (void)setDoubleValue:(double)val
{
    val = MAX(MIN(val, mMaxValue), mMinValue);
    super.doubleValue = val;
    self.stepper.doubleValue = val;
    [self updateText];
}

// NSTextFieldDelegate
- (void)controlTextDidEndEditing:(NSNotification *)obj
{
    self.doubleValue = self.text.doubleValue;
    if (self.target != nil && self.action != nil) {
        [self sendAction:self.action to:self.target];
    }
}
@end
