//
//  Utils.m
//  UpTop
//
//  Created by Geoff Prewett on 11/19/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "Utils.h"

#import "Settings.h"

CGFloat calcEmPx()
{
    CGFloat emPx = 0;
    if (emPx == 0) {
        NSTextField *tf = createLabel(@"Ag");
        [tf sizeToFit];
        emPx = tf.frame.size.height;
    }
    return emPx;
}

NSTextField* createLabel(NSString *text)
{
    NSTextField *label = [NSTextField labelWithString:text];
    label.textColor = [NSColor whiteColor];
    label.font = Settings.font;
    // This seems to really slow things down a lot (from 3% CPU usage to 10%)
    //    [label sizeToFit];
    return label;
}

double toMB(uint64_t nBytes)
{
    return nBytes / 1048576.0;
}

NSString* formattedByteCount(uint64_t bytes)
{
    double kb = (double)bytes / 1024;
    if (kb > 0 && kb < 1) {
        return @"1 kB";
    }
    if (kb < 1000) {
        return [NSString stringWithFormat:@"%d kB", (int)round(kb)];
    }
    double mb = kb / 1024;
    if (mb < 1000) {
        return [NSString stringWithFormat:@"%.1f MB", mb];
    }
    return [NSString stringWithFormat:@"%.1f GB", mb / 1024.0];
}

CGFloat position(NSView *v, CGFloat x, CGFloat y, CGFloat w, CGFloat h)
{
    y -= h;
    v.frame = NSMakeRect(x, y, w, h);
    return y;
}

CGFloat positionXY(NSView *v, CGFloat x, CGFloat y)
{
    return position(v, x, y, v.frame.size.width, v.frame.size.height);
}

CGFloat placeRow(CGFloat x, CGFloat y, CGFloat w, CGFloat spacing, NSArray *views)
{
    int nNulls = 0;
    int nViews = 0;
    CGFloat totalWidth = 0.0;
    CGFloat maxHeight = 0.0;
    for (NSObject *o in views) {
        if (o == [NSNull null]) {
            nNulls++;
        } else {
            NSView *v = (NSView*)o;
            nViews++;
            totalWidth += v.frame.size.width;
            maxHeight = MAX(maxHeight, v.frame.size.height);
        }
    }
    totalWidth += (CGFloat)nViews * spacing;
    
    int n = 0;
    CGFloat nullWidth = MAX(0, (w - totalWidth) / (CGFloat)nNulls);
    for (NSObject *o in views) {
        if (o == [NSNull null]) {
            x += nullWidth;
            if (n > 0) {  // compensate for the extra space that will be added
                x -= spacing;
            }
        } else {
            NSView *v = (NSView*)o;
            if (n > 0) {
                x += spacing;
            }
            positionXY(v, x, y - (maxHeight - v.frame.size.height) / 2.0);
            x += v.frame.size.width;
            n++;
        }
    }
    return maxHeight;
}
